@@ -34,6 +34,28 @@ def task(*args, **kwargs):
 
     return wrapper if invoked else wrapper(func)
 
+def _wrap_as_new(original, new):
+    if isinstance(original, tasks.Task):
+        return tasks.WrappedCallableTask(new)
+    return new
+
+
+def _list_annotating_decorator(attribute, *values):
+    def attach_list(func):
+        @wraps(func)
+        def inner_decorator(*args, **kwargs):
+            return func(*args, **kwargs)
+        _values = values
+        # Allow for single iterable argument as well as *args
+        if len(_values) == 1 and not isinstance(_values[0], basestring):
+            _values = _values[0]
+        setattr(inner_decorator, attribute, list(_values))
+        # Don't replace @task new-style task objects with inner_decorator by
+        # itself -- wrap in a new Task object first.
+        inner_decorator = _wrap_as_new(func, inner_decorator)
+        return inner_decorator
+    return attach_list
+
 
 def hosts(*host_list):
     """
@@ -58,18 +80,7 @@ def hosts(*host_list):
         Allow a single, iterable argument (``@hosts(iterable)``) to be used
         instead of requiring ``@hosts(*iterable)``.
     """
-
-    def attach_hosts(func):
-        @wraps(func)
-        def inner_decorator(*args, **kwargs):
-            return func(*args, **kwargs)
-        _hosts = host_list
-        # Allow for single iterable argument as well as *args
-        if len(_hosts) == 1 and not isinstance(_hosts[0], basestring):
-            _hosts = _hosts[0]
-        inner_decorator.hosts = list(_hosts)
-        return inner_decorator
-    return attach_hosts
+    return _list_annotating_decorator('hosts', *host_list)
 
 
 def roles(*role_list):
@@ -99,17 +110,7 @@ def roles(*role_list):
         Allow a single, iterable argument to be used (same as
         `~fabric.decorators.hosts`).
     """
-    def attach_roles(func):
-        @wraps(func)
-        def inner_decorator(*args, **kwargs):
-            return func(*args, **kwargs)
-        _roles = role_list
-        # Allow for single iterable argument as well as *args
-        if len(_roles) == 1 and not isinstance(_roles[0], basestring):
-            _roles = _roles[0]
-        inner_decorator.roles = list(_roles)
-        return inner_decorator
-    return attach_roles
+    return _list_annotating_decorator('roles', *role_list)
 
 
 def runs_once(func):
@@ -136,6 +137,7 @@ def runs_once(func):
         if not hasattr(decorated, 'return_value'):
             decorated.return_value = func(*args, **kwargs)
         return decorated.return_value
+    decorated = _wrap_as_new(func, decorated)
     # Mark as serial (disables parallelism) and return
     return serial(decorated)
 
@@ -153,7 +155,7 @@ def serial(func):
     """
     if not getattr(func, 'parallel', False):
         func.serial = True
-    return func
+    return _wrap_as_new(func, func)
 
 
 def parallel(pool_size=None):
@@ -177,7 +179,7 @@ def parallel(pool_size=None):
         inner.parallel = True
         inner.serial = False
         inner.pool_size = pool_size
-        return inner
+        return _wrap_as_new(func, inner)
 
     # Allow non-factory-style decorator use (@decorator vs @decorator())
     if type(pool_size) == type(real_decorator):
@@ -209,5 +211,5 @@ def with_settings(**kw_settings):
         def inner(*args, **kwargs):
             with settings(**kw_settings):
                 return func(*args, **kwargs)
-        return inner
+        return _wrap_as_new(func, inner)
     return outer
